package anywheresoftware.b4j.objects;

import com.lynden.gmapsfx.GoogleMapView;
import com.lynden.gmapsfx.MapComponentInitializedListener;
import com.lynden.gmapsfx.javascript.event.MapStateEventType;
import com.lynden.gmapsfx.javascript.event.StateEventHandler;
import com.lynden.gmapsfx.javascript.event.UIEventHandler;
import com.lynden.gmapsfx.javascript.event.UIEventType;
import com.lynden.gmapsfx.javascript.object.GoogleMap;
import com.lynden.gmapsfx.javascript.object.LatLong;
import com.lynden.gmapsfx.javascript.object.LatLongBounds;
import com.lynden.gmapsfx.javascript.object.MVCArray;
import com.lynden.gmapsfx.javascript.object.MapOptions;
import com.lynden.gmapsfx.javascript.object.MapTypeIdEnum;
import com.lynden.gmapsfx.javascript.object.Marker;
import com.lynden.gmapsfx.javascript.object.MarkerOptions;
import com.lynden.gmapsfx.shapes.Polygon;
import com.lynden.gmapsfx.shapes.PolygonOptions;
import com.lynden.gmapsfx.shapes.Polyline;
import com.lynden.gmapsfx.shapes.PolylineOptions;

import anywheresoftware.b4a.AbsObjectWrapper;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.Events;
import anywheresoftware.b4a.BA.Hide;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.objects.collections.List;
import anywheresoftware.b4j.objects.JFX.Colors;
import anywheresoftware.b4j.objects.PaneWrapper.ConcretePaneWrapper;
import javafx.scene.paint.Paint;
import netscape.javascript.JSObject;

/**
 *Uses GoogleMap JavaScript V3 API to show a map inside WebView.
 *GoogleMap.AsPane returns a Pane which should be added to the nodes tree.
 *The map is only ready after Ready event is raised.
 *Requires Java 8+.
 */
@ShortName("GoogleMap")
@Events(values={
		"Ready", 
		"CameraChange (Position As CameraPosition)",
		"Click (Point As LatLng)",
		"MarkerClick (SelectedMarker As Marker)"
})
@Version(1.26f)
public class GoogleMapWrapper {
	@Hide
	public GoogleMap map;
	@Hide
	public GoogleMapView mapView;
	private BA ba;
	private String eventName;
	private boolean isReady;
	public static final MapTypeIdEnum MAP_TYPE_NORMAL = MapTypeIdEnum.ROADMAP;
	public static final MapTypeIdEnum MAP_TYPE_SATELLITE = MapTypeIdEnum.SATELLITE;
	public static final MapTypeIdEnum MAP_TYPE_TERRAIN = MapTypeIdEnum.TERRAIN;
	public static final MapTypeIdEnum MAP_TYPE_HYBRID = MapTypeIdEnum.HYBRID;

	/**
	 * Initializes the map. The Ready event will be raised when the map is ready.
	 *EventName - Sets the subs that will handle the events.
	 *Options - MapOptions object. Pass Null for the default options.
	 */
	public void Initialize(BA ba, String EventName, MapOptionsWrapper Options) {
		mapView = new GoogleMapView();
		this.eventName = EventName.toLowerCase(BA.cul);
		this.ba = ba;
		mapView.addMapInializedListener(new MapComponentInitializedListener() {

			@Override
			public void mapInitialized() {
				isReady = false;
				MapOptions mapOptions;
				mapOptions = new MapOptions();
				mapOptions.center(new LatLong(20, 20)).zoom(2);
				if (Options != null) {
					mapOptions.streetViewControl(Options.StreetViewControl).overviewMapControl(Options.OverviewMapControl)
					.panControl(Options.PanControl).zoomControl(Options.ZoomControl)
					.mapTypeControl(Options.MapTypeControl);
				}

				map = mapView.createMap(mapOptions);
				mapOptions.mapType(MapTypeIdEnum.SATELLITE);
				if (ba.subExists(eventName + "_camerachange")) {
					map.addStateEventHandler(MapStateEventType.bounds_changed, new StateEventHandler() {

						@Override
						public void handle() {
							ba.raiseEventFromUI(GoogleMapWrapper.this, eventName + "_camerachange", getCameraPosition());
						}
					});
				}
				isReady = true;
				ba.raiseEvent(GoogleMapWrapper.this, eventName + "_ready");
				if (ba.subExists(eventName + "_click")) {
					map.addUIEventHandler(UIEventType.click, new UIEventHandler() {

						@Override
						public void handle(JSObject obj) {
							Object o = obj.getMember("latLng");
							if (o instanceof JSObject) {
								JSObject latLng = (JSObject)o;
								ba.raiseEventFromUI(GoogleMapWrapper.this, eventName + "_click", 
										AbsObjectWrapper.ConvertToWrapper(new LatLngWrapper(), new LatLong(latLng)));
							}
						}
					});
				}
			}
		});
	}
	/**
	 * Moves the camera position.
	 */
	public void MoveCamera(CameraPositionWrapper NewPosition) {
		map.setCenter(NewPosition.ll);
		map.setZoom(NewPosition.zoom);
	}
	/**
	 * Returns the current camera position.
	 */
	public CameraPositionWrapper getCameraPosition() {
		CameraPositionWrapper cpw = new CameraPositionWrapper();
		cpw.ll = map.getCenter();
		cpw.zoom = map.getZoom();
		return cpw;
	}
	/**
	 * Adds a marker to the map with a custom icon.
	 *The IconUri must point to an online image or an image from the assets folder.
	 *<b>In the later case the custom icon will only appear in Release mode.</b>
	 *Example: <code>
	 * gmap.AddMarker2(10, 10, "This is a test", _
	 *	 	File.GetUri(File.DirAssets, "SomeIcon.png"))</code>
	 */
	public MarkerWrapper AddMarker2(double Lat, double Lon, String Title, String IconUri) {
		MarkerOptions mo = new MarkerOptions().position(new LatLong(Lat, Lon))
				.title(Title);
		if (IconUri.length() > 0) {
			if (!IconUri.startsWith("file") || BA.isShellModeRuntimeCheck(ba) == false)
				mo.icon(IconUri);
		}
		final Marker m = new Marker(mo);
		map.addMarker(m);
		if (ba.subExists(eventName + "_markerclick")) {
			map.addUIEventHandler(m, UIEventType.click, new UIEventHandler() {

				@Override
				public void handle(JSObject obj) {
					ba.raiseEventFromUI(GoogleMapWrapper.this, eventName + "_markerclick", 
							(MarkerWrapper)AbsObjectWrapper.ConvertToWrapper(new MarkerWrapper(), m));
				}
			});
		}
		return (MarkerWrapper)AbsObjectWrapper.ConvertToWrapper(new MarkerWrapper(), m);
	}
	/**
	 * Adds a marker to the map.
	 */
	public MarkerWrapper AddMarker(double Lat, double Lon, String Title) {
		return AddMarker2(Lat, Lon, Title, "");
	}
	/**
	 * Adds a polyline to the map.
	 *Points - A list or array of LatLng points.
	 *Width - Line width.
	 *Color - Line color.
	 */
	public PolylineWrapper AddPolyline(List Points, float Width, Paint Color) {
		PolylineOptions po = new PolylineOptions();
		po.strokeWeight(Width);
		po.strokeColor(htmlColor(Color));
		po.path(new MVCArray(Points.getObject().toArray()));
		Polyline line = new Polyline(po);
		map.addMapShape(line);
		return (PolylineWrapper)AbsObjectWrapper.ConvertToWrapper(new PolylineWrapper(), line);
	}
	/** Removes the specified polyline from the map. */
	public void RemovePolyline(PolylineWrapper Polyline) {
		map.removeMapShape(Polyline.getObject());
	}

	/**
	 * Adds a polygon to the map.
	 *Points - A list or array of LatLng points.
	 *StrokeWidth - Stroke width.
	 *StrokeColor - Stroke color.
	 *FillColor - Inner color.
	 *Opacity - Inner color opacity. Value between 0 to 1. 
	 */
	public PolygonWrapper AddPolygon(List Points, float StrokeWidth, Paint StrokeColor, Paint FillColor, double Opacity) {
		PolygonOptions po = new PolygonOptions();
		po.strokeWeight(StrokeWidth);
		po.fillColor(htmlColor(FillColor));
		po.strokeColor(htmlColor(StrokeColor));
		po.fillOpacity(Opacity);
		po.paths(new MVCArray(Points.getObject().toArray()));
		Polygon gon = new Polygon(po);
		map.addMapShape(gon);
		return (PolygonWrapper)AbsObjectWrapper.ConvertToWrapper(new PolygonWrapper(), gon);

	}
	/** Removes the specified polygon from the map. */
	public void RemovePolygon(PolygonWrapper Polygon) {
		map.removeMapShape(Polygon.getObject());
	}

	@Hide
	public String htmlColor(Paint Color) {
		return String.format("#%06X", (0xFFFFFF & Colors.To32Bit(Color)));
	}

	/**
	 * Sets the map type. The value should be one of the MAP_TYPE constants. 
	 */
	public void setMapType(Object v) {
		map.setMapType((MapTypeIdEnum) v);
	}
	/**
	 * Returns the pane that holds the map.
	 */
	public ConcretePaneWrapper getAsPane() {
		return (ConcretePaneWrapper) AbsObjectWrapper.ConvertToWrapper(new ConcretePaneWrapper(), mapView);
	}


	/**
	 * The map view is modeled as a camera looking down on a flat plane.
	 *See this <link>link|https://developers.google.com/maps/documentation/android/views#the_camera_position</link> for more information about the possible values.
	 */
	@ShortName("CameraPosition")
	public static class CameraPositionWrapper  {
		LatLong ll;
		int zoom;
		/**
		 * Initializes the camera position with the given latitude, longitude and zoom.
		 */
		public void Initialize(double Lat, double Lng, int Zoom) {
			ll = new LatLong(Lat, Lng);
			zoom = Zoom;
		}

		/**
		 * Returns the location that the camera is pointing at.
		 */
		public LatLngWrapper getTarget() {
			return (LatLngWrapper)AbsObjectWrapper.ConvertToWrapper(new LatLngWrapper(), ll);
		}

		/**
		 * Returns the zoom level.
		 */
		public float getZoom() {
			return zoom;
		}
		@Override
		public String toString() {
			return String.valueOf(ll) + ", zoom: " + zoom;
		}
	}

	/**
	 * Holds latitude and longitude values.
	 */
	@ShortName("LatLng")
	public static class LatLngWrapper extends AbsObjectWrapper<LatLong> {
		/**
		 * Returns the latitude value.
		 */
		public double getLatitude() {
			return getObject().getLatitude();
		}
		/**
		 * Returns the longitude value.
		 */
		public double getLongitude() {
			return getObject().getLongitude();
		}
		/**
		 * Initializes a new object.
		 */
		public void Initialize(double Latitude, double Longitude) {
			setObject(new LatLong(Latitude, Longitude));
		}
	}

	@ShortName("Marker")
	public static class MarkerWrapper extends AbsObjectWrapper<Marker> {
		/**
		 * Sets the marker position.
		 */

		public void setPosition(LatLngWrapper value) {
			getObject().setPosition(value.getObject());
		}
		/**
		 * Gets or sets whether the marker is visible.
		 */
		public void setVisible(boolean v) {
			getObject().setVisible(v);
		}
		public boolean getVisible() {
			return getObject().getVisible();
		}
		/**
		 * Removes the marker from the map.
		 */
		public void Remove() {
			getObject().setMap(null);
		}
		/**
		 * Sets the marker title.
		 */
		public void setTitle(String v) {
			getObject().setTitle(v);
		}

	}
	@ShortName("Polyline")
	public static class PolylineWrapper extends AbsObjectWrapper<Polyline> {
		/**
		 * Gets or sets whether this line is visible.
		 */
		public boolean getVisible() {
			return getObject().getVisible();
		}
		public void setVisible(boolean v) {
			getObject().setVisible(v);
		}
	}
	@ShortName("Polygon")
	public static class PolygonWrapper extends AbsObjectWrapper<Polygon> {
		/**
		 * Gets or sets whether this line is visible.
		 */
		public boolean getVisible() {
			return getObject().getVisible();
		}
		public void setVisible(boolean v) {
			getObject().setVisible(v);
		}
	}
	/**
	 * MapOptions is passed to GoogleMap.Initialize method.
	 *It allows you to choose which controls will be available.
	 */
	@ShortName("MapOptions")
	public static class MapOptionsWrapper  {
		public boolean PanControl = true;
		public boolean StreetViewControl = true;
		public boolean OverviewMapControl = true;
		public boolean MapTypeControl = true;
		public boolean ZoomControl = true;
	}

	@BA.ShortName("LatLongBounds")
	public static class LatLongBoundsWrapper
	extends AbsObjectWrapper<LatLongBounds>
	{
		public LatLngWrapper getNorthEast()
		{
			return (LatLngWrapper)AbsObjectWrapper.ConvertToWrapper(new LatLngWrapper(), getObject().getNorthEast());
		}

		public LatLngWrapper getSouthWest()
		{
			return (LatLngWrapper)AbsObjectWrapper.ConvertToWrapper(new LatLngWrapper(), getObject().getSouthWest());
		}

		public void Initialize(LatLngWrapper NorthEast, LatLngWrapper SouthWest)
		{
			setObject(new LatLongBounds(SouthWest.getObject(), NorthEast.getObject()));
		}
	}

	/**
	 *Returns the LatLongBounds of the visual area. It is recommended to read Bounds only inside the CameraChange event because its values may be incorrect otherwise.
	 */
	public LatLongBoundsWrapper getBounds() {
		return (LatLongBoundsWrapper)AbsObjectWrapper.ConvertToWrapper(new LatLongBoundsWrapper(), map.getBounds());
	}

	/**Moves the map to ensure the given bounds fit within the viewport.
	 *Note that the Google Maps API will add a margin around these bounds.
	 */
	public void FitBounds(LatLongBoundsWrapper bounds ) {
		map.fitBounds(bounds.getObject());
	}

	/** Pans the map by the minimum amount necessary to contain the given LatLongBounds. It makes no guarantee where on the map the bounds will be, except that as much of the bounds as possible will be visible. */
	public void PanToBounds(LatLongBoundsWrapper bounds) {
		map.panToBounds(bounds.getObject());
	} 

	/** Sets the zoom level. */
	public void setZoom(int zoom) {
		map.setZoom(zoom);
	}

	/** Returns True when the map is ready. */
	public boolean IsReady()
	{
		return isReady;
	}
}
